"""
Logic for downloading input files from URIs
"""
import os
from typing import Optional, List, Iterable, Dict

# WDL tasks for downloading a file based on its URI scheme
# TODO: formalize plugin mechanism

_downloaders = {}
_downloaders[
    "https"
] = r"""
task aria2c {
    input {
        String uri
        Int connections = 10
    }
 
    command <<<
        set -euxo pipefail
        mkdir __out
        cd __out
        aria2c -x ~{connections} -s ~{connections} \
            --file-allocation=none --retry-wait=2 --stderr=true --enable-color=false \
            "~{uri}"
    >>>

    output {
        File file = glob("__out/*")[0]
    }

    runtime {
        cpu: 4
        memory: "1G"
        docker: "hobbsau/aria2"
    }
}
"""
_downloaders["http"] = _downloaders["https"]
_downloaders["ftp"] = _downloaders["https"]

_downloaders[
    "gs"
] = r"""
task gsutil_cp {
    input {
        String uri
    }
    command <<<
        set -euxo pipefail
        mkdir __out/
        gsutil -q cp "~{uri}" __out/
    >>>
    output {
        File file = glob("__out/*")[0]
    }
    runtime {
        cpu: 4
        memory: "1G"
        docker: "google/cloud-sdk:slim"
    }
}
"""


def _downloader(uri: str) -> Optional[str]:
    colon = uri.find(":")
    if colon <= 0:
        return None
    scheme = uri[:colon]
    return _downloaders.get(scheme, None)


def schemes() -> Iterable[str]:
    yield from _downloaders.keys()


def able(uri: str) -> bool:
    """
    Returns True if uri appears to be a URI we know how to download
    """
    return _downloader(uri) is not None


def run(uri: str, **kwargs) -> str:
    """
    Download the URI and return the local filename.

    kwargs are passed through to ``run_local_task``, so ``run_dir`` and ``logger_prefix`` may be
    useful in particular.
    """
    from . import run_local_task, RunFailed, DownloadFailed
    from .. import parse_tasks, values_from_json

    downloader_wdl = _downloader(uri)
    assert downloader_wdl

    task = parse_tasks(downloader_wdl, version="1.0")[0]  # pyre-ignore
    task.typecheck()
    inputs = values_from_json({"uri": uri}, task.available_inputs)  # pyre-ignore
    try:
        subdir, outputs = run_local_task(task, inputs, **kwargs)
    except RunFailed as exn:
        raise DownloadFailed(uri) from exn.__cause__
    return outputs["file"].value



# pyre-strict
from typing import Union, Dict, Tuple, Any
from .. import Tree, Value, Env
from . import task
from . import workflow
from .error import *
from .task import run_local_task, make_output_links
from .workflow import run_local_workflow


def run(
    exe: Union[Tree.Task, Tree.Workflow],
    inputs: Env.Bindings[Value.Base],
    **run_kwargs: Dict[str, Any],
) -> Tuple[str, Env.Bindings[Value.Base]]:
    """
    Run the task or workflow given the inputs environment, and any configuration arguments to
    ``run_local_{task,workflow}``.

    ``inputs`` may be parsed from a JSON dict using :func:`~WDL.values_from_json`. The
    workflow/task name should NOT be used as a namespace for the input values. They should have
    been typechecked against ``exe.available_inputs`` already.
    """
    if "max_tasks" in run_kwargs and isinstance(exe, Tree.Task):
        del run_kwargs["max_tasks"]  # N/A to run_local_task
    entrypoint = run_local_task if isinstance(exe, Tree.Task) else run_local_workflow
    return entrypoint(exe, inputs, **run_kwargs)  # pyre-ignore



# pyre-strict
from typing import Union
from ..Error import RuntimeError as _RuntimeError
from ..Tree import Task as _Task, Workflow as _Workflow


class CommandFailed(_RuntimeError):
    """
    Failure of the task command
    """

    exit_status: int
    """
    Task command exit status
    """

    stderr_file: str
    """
    Path to a file containing the task's standard error
    """

    def __init__(self, exit_status: int, stderr_file: str, message: str = "") -> None:
        super().__init__(message or f"task command failed with exit status {exit_status}")
        self.exit_status = exit_status
        self.stderr_file = stderr_file


class Terminated(_RuntimeError):
    """
    Workflow/task was terminated, e.g. by Unix signal
    """

    pass


class OutputError(_RuntimeError):
    """
    Failure whilst gathering task outputs
    """

    pass


class DownloadFailed(_RuntimeError):
    """
    Failure to download a URI input file
    """

    uri: str

    def __init__(self, uri: str, message: str = "") -> None:
        super().__init__(message or ("unable to download " + uri))
        self.uri = uri


class RunFailed(_RuntimeError):
    """

    """

    exe: Union[_Task, _Workflow]
    run_id: str
    run_dir: str

    def __init__(self, exe: Union[_Task, _Workflow], run_id: str, run_dir: str) -> None:
        super().__init__(
            f"{'task' if  isinstance(exe, _Task) else 'workflow'} {exe.name} ({exe.pos.uri} Ln {exe.pos.line} Col {exe.pos.column}) failed"
        )
        self.exe = exe
        self.run_id = run_id
        self.run_dir = run_dir
